import os

CACHE_REDIS_HOST = os.environ.get("CACHE_REDIS_HOST")
CACHE_REDIS_PORT = os.environ.get("CACHE_REDIS_PORT")
CACHE_REDIS_DB = os.environ.get("CACHE_REDIS_DB")
CACHE_REDIS_PASSWORD = os.environ.get("CACHE_REDIS_PASSWORD")
CACHE_NAMESPACE = os.environ.get("CACHE_NAMESPACE")
